from __future__ import annotations

import logging

import pandas as pd
from mltraq.utils.exceptions import FatalError
from tqdm.auto import tqdm

from cumulative.opts import options
from cumulative.utils.frames import drop_cols_with_prefix

log = logging.getLogger(__name__)


def process_row(func, row, **kwargs):
    return func(row, **kwargs)


class Transform:
    """
    Abstraction of transforms.
    """

    def __init__(self, c, name: str = None):
        """
        Link transform to Cumulative class and set its name.
        """

        self.c = c
        self.name = self.__class__.__name__.lower() if not name else name

    def transform_row(self, row: pd.Series) -> pd.Series:
        """
        Transformations can be applied either on rows or columns.
        For rows, the `transform_row` method is overloaded and it
        returns a series with the new columns to add.
        """
        return pd.Series()

    def __call__(self, **kwargs):
        """
        Handles transform requests.
        """
        tqdm.pandas(**(options().get("tqdm") | {"desc": self.name}))

        # Ensure src and dst of operation is set
        kwargs["src"] = options().get("transforms.src", prefer=kwargs.pop("src", None))
        kwargs["dst"] = options().get("transforms.dst", prefer=kwargs.pop("dst", None))
        kwargs["drop"] = options().get("transforms.drop", prefer=kwargs.pop("drop", None))

        # Apply transform and obtain new columns
        kwargs_apply = kwargs.copy()
        dst = kwargs_apply.pop("dst")
        kwargs_apply.pop("drop")
        # dst, drop parameters are removed before executing .apply, as they are handled
        # transparently in this method.
        df = self.apply(**kwargs_apply)

        # Log operation
        self.c.lineage.track(self.name, dst, kwargs)

        if df is None:
            # No change (e.g., sorting)
            return self.c
        elif isinstance(df, pd.DataFrame):
            # Multiple columns added from processing entire data frame
            df.columns = [f"{kwargs['dst']}.{col}" if col != "idx" else col for col in df.columns]
        elif isinstance(df, pd.Series):
            # Multiple columns added from processing single row
            df = pd.DataFrame({kwargs["dst"]: df})
        else:
            raise FatalError("Invalid transform result; expected None, pandas.Series or pandas.DataFrame")

        if kwargs["drop"]:
            # If drop specified, clean up prefix `dst` before adding the new columns.
            # This ensures that we don't mix up results of different transforms in the same prefix.
            self.c.df = drop_cols_with_prefix(self.c.df, kwargs["dst"])

        self.c.df = pd.concat(
            [self.c.df, df],
            axis=1,
        )
        return self.c

    def apply(self, **kwargs):
        """
        Transformations can be applied either on rows or columns.
        For columns, the `apply` method is overloaded and it
        returns a data frame with the new columns to add.

        By default, it handles row transforms.
        """
        return self.c.df.progress_apply(lambda row: self.transform_row(row, **kwargs), axis=1)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform
from cumulative.utils.frames import rows_with_prefix


class Apply(Transform):
    def transform_row(self, row: pd.Series, src: str, func: callable = None, **kwargs) -> pd.Series:
        """
        Apply `func` to `row`, with `src` column prefix omitted to increase reusability.
        """

        row = row[rows_with_prefix(row, src)]
        row.index = [col.removeprefix(f"{src}.") for col in row.index]

        return pd.Series(func(row, **kwargs))


def start_zero(row: pd.Series) -> pd.Series:
    """
    Insert [0,0] as first pair for series x,y
    """

    row["x"] = np.insert(row["x"], 0, 0)
    row["y"] = np.insert(row["y"], 0, 0)
    return row


def last_x_nonzero_y(row: pd.Series, col: str = "value") -> pd.Series:
    """
    Sets `col` to the last value in x with a nonzero value on y,
    useful to trim ending zeros on y.
    """

    row[col] = row["x"][np.flatnonzero(row["y"])[-1]]
    return row


def limit_region(
    row: pd.Series, x_min: float = None, x_max: float = None, y_min: float = None, y_max: float = None
) -> pd.Series:
    """
    Limit x,y to a region [x_min,x_max] and [y_min,y_max]
    """

    if isinstance(x_min, str):
        x_min = row[x_min]
    if isinstance(x_max, str):
        x_max = row[x_max]
    if isinstance(y_min, str):
        y_min = row[y_min]
    if isinstance(y_max, str):
        y_max = row[y_max]

    m = np.ones_like(row["x"], dtype=bool)

    if x_min:
        m &= row["x"] >= x_min

    if x_max:
        m &= row["x"] <= x_max

    if y_min:
        m &= row["y"] >= y_min

    if y_max:
        m &= row["y"] <= y_max

    row["x"] = row["x"][m]
    row["y"] = row["y"][m]

    return row



import numpy as np
import pandas as pd
from sklearn.datasets import load_iris

from cumulative import Cumulative


def load_sklearn(func: callable = load_iris) -> Cumulative:
    """
    Load a scikit-learn dataset by passing the loading function `func`.
    """

    data, target = func(return_X_y=True)
    df = pd.DataFrame(data).apply(
        lambda row: pd.Series({"base.x": np.arange(row.values.shape[0]), "base.y": row.values}), axis=1
    )
    df["base.z"] = target
    c = Cumulative(df)
    c.meta.name = func.__name__

    return c



import pandas as pd

from cumulative import Cumulative
from cumulative.loaders.preprocess import melt, nest


def load_wide() -> Cumulative:
    """
    Load a dataset of time series in wide format with numerical and categorical columns.
    """

    df = pd.DataFrame(
        {
            "id": ["S1", "S2", "S3"],
            "category": ["A", "A", "B"],
            "t1": [5, 10, 5],
            "t2": [10, 15, 10],
            "t3": [30, 35, 15],
            "t4": [50, 55, 20],
            "t5": [20, 60, 25],
            "rate": [5, 2, 8],
        }
    )

    df = melt(df, group="id", attributes=["category", "rate"])
    df = nest(df)

    c = Cumulative(df)

    c.lineage.track("load_wide")

    return c



from cumulative.transforms.transform import Transform
from cumulative.utils.frames import drop_cols_with_prefix


class Drop(Transform):
    def apply(self, src: str):
        """
        Drop columns with prefix `src`
        """
        self.c.df = drop_cols_with_prefix(self.c.df, src)



import re

import numpy as np
import pandas as pd

from cumulative.opts import options


def melt(
    df: pd.DataFrame,
    group: str,
    pattern: str = "^t(.*)$",
    x_dtype: type = None,
    y_dtype: type = None,
    attributes: list[str] | None = None,
) -> pd.DataFrame:
    """
    Unpivot dataframe `dt` from wide to long format. Similar to Pandas.melt, but more flexible.
    `pattern` controls which columns provide numerical values, and instructs how to extract them.
    `attributes` lists the columns to be retained as attributes of the series (not necessarily numeric).
    """

    if attributes is None:
        attributes = []

    if x_dtype is None:
        x_dtype = np.dtype("float64")

    if y_dtype is None:
        y_dtype = np.dtype("float64")

    # example: df has columns ['id', 'category', 't1', 't2', 't3', 't4', 't5']
    # extract list of pairs (col_name, matching_substring) from all columns with at least one match
    matches = [
        (cm[0], cm[1].group(1)) for cm in [(c, re.search(pattern, c, re.IGNORECASE)) for c in df.columns] if cm[1]
    ]

    def f(row: pd.Series) -> pd.Series:
        # Extract a series form a row, with arrays representing the values and the attributes.
        # Order of columns is important. eg., V[5] comes after V[4].
        C = [match[0] for match in matches]  # example: "t2"
        T = [match[1] for match in matches]  # example: 2
        V = [row[match[0]] for match in matches]  # example: 12.44
        return pd.Series(
            {"c": C, "x": T, "y": V} | {"name": row[group]} | {f"attr.{attr}": row[attr] for attr in attributes}
        )

    df = df.apply(f, axis=1).explode(["c", "x", "y"])

    df["x"] = df["x"].astype(x_dtype)
    df["y"] = df["y"].astype(y_dtype)

    # example: {'col': 't1', 'time': 1, 'value': 5, 'category: 'A'
    return df


def nest(df: pd.DataFrame, dst=None) -> pd.DataFrame:
    """
    Transform series in long format to nested format. Expected columns:
    - "name": name of the series
    - "c": value of the columns representing x
    - "x": extracted x value from "c"
    - "y": value at position "x"
    - "attr.*": attributes associated to the series (optional)
    """

    dst = options().get("transforms.dst", prefer=dst)

    attributes = [c for c in df.columns if c.startswith("attr.")]

    df = df.groupby("name").agg({"c": list, "x": list, "y": list} | {attr: "first" for attr in attributes})
    df["x"] = df["x"].apply(lambda V: np.array(V))
    df["y"] = df["y"].apply(lambda V: np.array(V))

    df = df.reset_index(names="name")
    df = df.rename(
        columns={"name": f"{dst}.name", "c": f"{dst}.c", "x": f"{dst}.x", "y": f"{dst}.y"}
        | {attr: f"{dst}.{attr}" for attr in attributes}
    )

    return df


def nest_xy(df: pd.DataFrame, name: str = "name", y: str = "y") -> pd.DataFrame:
    """
    Transform series in long format to nested format. Expected columns:
    - `name`: name of the series
    - `y`: value at a certain offset

    The X column will be deduced by the offset of Y.
    """

    df = df[[name, y]].rename(columns={name: "name", y: "base.y"})
    df["base.x"] = df.groupby("name").cumcount()
    df = df.groupby("name").agg({"base.x": list, "base.y": list}).reset_index()
    df["base.x"] = df["base.x"].apply(lambda a: np.array(a))
    df["base.y"] = df["base.y"].apply(lambda a: np.array(a))
    df = df[["name", "base.x", "base.y"]]
    return df



import numpy as np
import pandas as pd
from scipy.stats import rv_continuous

from cumulative import Cumulative
from cumulative.opts import options
from cumulative.utils.lineage import get_kwargs


class ExceptionInvalidKind(Exception):
    pass


def dist_record_cdf(dist: rv_continuous, num: int, dst: str) -> dict:
    """
    Generate a new record representing the `dist` distribution CDF
    with `num` points and `dst` key prefix.
    """

    x = np.linspace(dist.ppf(0.01), dist.ppf(0.99), num)
    y = dist.cdf(x)
    return {f"{dst}.x": x, f"{dst}.y": y}


def dist_record_pdf(dist: rv_continuous, num: int, dst: str) -> dict:
    """
    Generate a new record representing the `dist` distribution PDF
    with `num` points and `dst` key prefix.
    """

    x = np.linspace(dist.ppf(0.01), dist.ppf(0.99), num)
    y = dist.pdf(x)
    return {f"{dst}.x": x, f"{dst}.y": y}


def dist_record_rvs(dist: rv_continuous, num: int, dst: str) -> dict:
    """
    Generate a new record representing the `dist` distribution
    random samples with `num` points and `dst` key prefix.
    Points are sorted by ascending order.
    """

    x = np.linspace(0, 1, num)
    y = np.sort(dist.rvs(size=num))
    return {f"{dst}.x": x, f"{dst}.y": y}


def load_dist(dists: list[rv_continuous], kind: str = "cdf", num: int = 100, dst: str = None) -> Cumulative:
    """
    Load a synthetic dataset representing `dists` curves, with CDF (kind="cdf") or random samples
    (kind="rvs"), each with `num` points, with prefix `dst`.
    """

    dst = options().get("transforms.dst", prefer=dst)

    records = []
    if kind == "cdf":
        for dist in dists:
            records.append(dist_record_cdf(dist, num, dst))
    elif kind == "rvs":
        for dist in dists:
            records.append(dist_record_rvs(dist, num, dst))
    elif kind == "pdf":
        for dist in dists:
            records.append(dist_record_pdf(dist, num, dst))
    else:
        raise ExceptionInvalidKind(f"No valid kind: '{kind}'")

    df = pd.DataFrame(records)
    c = Cumulative(df)

    c.lineage.track("load_dist", dst, get_kwargs())

    return c



from mltraq.utils.base_options import BaseOptions


class Options(BaseOptions):
    default_values = {
        "reproducibility": {"random_seed": 123},
        "tqdm": {"disable": False, "leave": False, "delay": 0},
        "transforms": {
            "src": "base",
            "dst": "base",
            "tmp": "temp",
            "drop": True,
        },
        "plot": {
            # Defaults used by collection visualizers
            "color": "black",
            "colormap": "cool",
            "ctx": {
                # Templates for mltraq.utils.plotting.plot_ctx(...) parameters
                "default": {
                    "spines_bottomleft": True,
                    "x_label": "X",
                    "y_label": "Y",
                    "figsize": (3, 3),
                    "facecolor": "white",
                },
                "whiteboard": {"show_axes": False, "figsize": (3, 3)},
            },
        },
        "warnings": {"disable": False},
        "doc": {"url": "https://elehcimd.github.io/cumulative/"},
    }


def options() -> BaseOptions:
    """
    Returns singleton object of options.
    """

    return Options.instance()



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform
from cumulative.utils.frames import drop_prefix


class ExceptionInvalidMethod(Exception):
    pass


class Score(Transform):
    def apply(self, src: str, func: callable = None, method: str = "value", reverse: bool = False) -> pd.DataFrame:
        """
        Extract a score from `src` column using `method`.
        Useful for sorting and setting colors.
        """

        if func:
            df = self.c.frame(src)
            df.columns = drop_prefix(src, df.columns)
            values = df.apply(func, axis=1)
        else:
            values = self.c.df[src].copy()
        if method == "value":
            values -= values.min()
            values /= values.max()
        elif method == "noscale":
            pass
        elif method == "argsort":
            values = np.argsort(values) / (values.shape[0] - 1)
        elif method == "index":
            values = np.linspace(0, 1, num=values.shape[0])
        else:
            raise ExceptionInvalidMethod(f"No valid method: '{method}'")

        if reverse:
            values = values[::-1]

        return pd.DataFrame({"value": values})



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cumulative',
 'cumulative.datasets',
 'cumulative.loaders',
 'cumulative.transforms',
 'cumulative.transforms.frame',
 'cumulative.transforms.row',
 'cumulative.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.1',
 'mltraq>=0.1.156',
 'pandas>=1.5.3',
 'scikit-learn>=1.2.2',
 'scipy>=1.11.4',
 'tqdm>=4.66.1']

setup_kwargs = {
    'name': 'cumulative',
    'version': '0.1.19',
    'description': 'Manipulate and Visualize Time Series Collections',
    'long_description': '<p align="center">\n<img width="75%" height="75%" src="https://elehcimd.github.io/cumulative/assets//img/logo-wide-black.svg" alt="Cumulative Logo">\n</p>\n\n<p align="center">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/test.svg" alt="Test">\n<img src="https://elehcimd.github.io/cumulative/assets//img/badges/coverage.svg" alt="Coverage">\n<img src="https://elehcimd.github.io/cumulative/assets//img/badges/python.svg" alt="Python">\n<img src="https://elehcimd.github.io/cumulative/assets//img/badges/pypi.svg" alt="PyPi">\n<img src="https://elehcimd.github.io/cumulative/assets//img/badges/license.svg" alt="License">\n<img src="https://elehcimd.github.io/cumulative/assets//img/badges/code-style.svg" alt="Code style">\n</p>\n\n---\n\n<h1 align="center">\nManipulate and Visualize Time Series Collections\n</h1>\n\nAn open-source Python library for Data Scientists to efficiently manipulate collections of time series data. Features include data loading, transformation, persistence, and visualization.\n\n---\n\n* **Documentation**: [https://elehcimd.github.io/cumulative](https://elehcimd.github.io/cumulative)\n* **Source code**: [https://github.com/elehcimd/cumulative](https://github.com/elehcimd/cumulative) (License: [BSD 3-Clause](https://elehcimd.github.io/cumulative/license/))\n* **Discussions**: [Ask questions, share ideas, engage](https://github.com/elehcimd/cumulative/discussions)\n* **Funding**: You can [star](https://github.com/elehcimd/cumulative) the project on GitHub and [hire me](https://www.linkedin.com/in/dallachiesa/) to make sense of your time series.\n\n---\n\n',
    'author': 'Michele Dallachiesa',
    'author_email': 'michele.dallachiesa@sigforge.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://elehcimd.github.io/cumulative/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.0',
}


setup(**setup_kwargs)



import pandas as pd

from cumulative.transforms.transform import Transform
from cumulative.utils.frames import columns_with_prefix, drop_prefix


class Copy(Transform):
    def apply(self, src: str) -> pd.DataFrame:
        """
        Copy columns with `src` prefix (destination handled in transforms.py)
        """
        cols = columns_with_prefix(self.c.df, src)
        df = self.c.df[cols].copy()
        df.columns = drop_prefix(src, cols)
        return df



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


def diff(values: np.ndarray, prefix: str) -> dict:
    """
    Return dictionary with diff of `values`, nested in key `prefix`.
    Inverse of .cumsum().
    """
    return {f"{prefix}": np.diff(values, prepend=0)}


class Diff(Transform):
    def transform_row(self, row: pd.Series, src: str, kind: str = "y") -> pd.Series:
        """
        Apply diff to `kind` dimensions in prefix `src`.
        """

        if kind == "xy":
            x = diff(row[f"{src}.x"], "x")
            y = diff(row[f"{src}.y"], "y")
        elif kind == "x":
            x = diff(row[f"{src}.x"], "x")
            y = {"y": row[f"{src}.y"]}
        elif kind == "y":
            x = {"x": row[f"{src}.x"]}
            y = diff(row[f"{src}.y"], "y")

        return pd.Series({**x, **y})
